/* GIO - GLib Input, Output and Streaming Library
 * 
 * Copyright (C) 2012 Colin Walters <walters@verbum.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GS_SUBPROCESS_CONTEXT_PRIVATE_H__
#define __GS_SUBPROCESS_CONTEXT_PRIVATE_H__

#include "gsystem-subprocess-context.h"

G_BEGIN_DECLS

struct _GSSubprocessContext
{
  GObject parent;

  GSpawnFlags flags;
  gchar **argv;
  gboolean has_argv0;
  char **envp;
  char *cwd;

  GSSubprocessStreamDisposition stdin_disposition;
  GSSubprocessStreamDisposition stdout_disposition;
  GSSubprocessStreamDisposition stderr_disposition;

  guint keep_descriptors : 1;
  guint search_path : 1;
  guint search_path_from_envp : 1;
  guint unused_flags : 29;

  gint stdin_fd;
  gchar *stdin_path;

  gint stdout_fd;
  gchar *stdout_path;

  gint stderr_fd;
  gchar *stderr_path;

  GArray *postfork_close_fds;
  GArray *inherit_fds;

  GSpawnChildSetupFunc child_setup_func;
  gpointer child_setup_data;
};

G_END_DECLS

#endif
